# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable
from pacman.model.placements import Placement


def make_missing_string(missing: Iterable[Placement]) -> str:
    """
    Unused likely to be removed.

    :param missing:
    :return: A String with the X, Y and P of each placement
    """
    return "; ".join(
        f"({placement.x}, {placement.y}, {placement.p})"
        for placement in missing)
